:- auto_table.

/* definition of rules */
supportive_rule(Name, Head, Body) :- strict(Name, Head, Body).
supportive_rule(Name, Head, Body) :- defeasible(Name, Head, Body).
supportive_rule(Name, Head, Body) :- strict0(Name, Head, Body). 
supportive_rule(Name, Head, Body) :- strict1(Name, Head, Body). 
supportive_rule(Name, Head, Body) :- strict2(Name, Head, Body). 
supportive_rule(Name, Head, Body) :- defeasible0(Name, Head, Body).
supportive_rule(Name, Head, Body) :- defeasible1(Name, Head, Body).
supportive_rule(Name, Head, Body) :- defeasible2(Name, Head, Body).
supportive_rule(Name, Head, Body) :- conf_defeasible(Name, Head, Body).
rule(Name,Head,Body) :- supportive_rule(Name, Head, Body).

/* definition of definitely */
definitely(X):- fact(X).
definitely(X):- strict(R,X,L), definitely_provable(L).
definitely(X):- strict0(R,X,L), definitely_provable(L).
definitely(neg(X)):- strict1(R,neg(X),L), definitely_provable(L), sk_not(definitely(X)).
definitely(X):- strict2(R,X,L), definitely_provable(L).
definitely_provable([]).
definitely_provable(X):- definitely(X).
definitely_provable([X1|X2]):- definitely_provable(X1), definitely_provable(X2).

/* definition of ambiguity blocking defeasibly */ 
defeasibly(X):- definitely(X).
defeasibly(X):- negation(X,X1), supportive_rule(R,X,L), defeasibly_provable(L), sk_not(definitely(X1)), sk_not(overruled(R,X)).
defeasibly_provable([]).
defeasibly_provable(X):- defeasibly(X).
defeasibly_provable([X1|X2]):- defeasibly_provable(X1), defeasibly_provable(X2).

overruled(R,X):- negation(X,X1), supportive_rule(S,X1,U), defeasibly_provable(U), sk_not(defeated(S,X1)).

defeated(S,X):- sup(T,S), negation(X,X1), supportive_rule(T,X1,V), defeasibly_provable(V).
defeated(S,X):- super(T,S), negation(X,X1), supportive_rule(T,X1,V), defeasibly_provable(V).

strict0(R(H),H,New):- strict(R,H,B),setof(~(Q),member(~(Q),B),D),minus_set(B,D,K),setof(neg(Q),member(~(Q),B),P),append(K,P,New).
strict1(R(neg(Q)),neg(Q),[]):- strict(R,H,B),member(~(Q),B).
strict2(R(negg(Q)),~(neg(Q)),Q):- strict(R,H,B),member(~(Q),B).

defeasible0(R(H),H,New):- defeasible(R,H,B),setof(~(Q),member(~(Q),B),D),minus_set(B,D,K),setof(neg(Q),member(~(Q),B),P),append(K,P,New).
defeasible1(R(neg(Q)),neg(Q),[]):- defeasible(R,H,B),member(~(Q),B). 
defeasible2(R(negg(Q)),~(neg(Q)),Q):- defeasible(R,H,B),member(~(Q),B).

conf_defeasible(R,Z,L):- strict(S,conflict,C), negation(Z,X), member(X,C), member(Y,C), defeasible(R,Y,L), sk_not(defeasible(R,X,L)).
super(R,S):- sup(R,S), defeasible(R,H1,L1), conf_defeasible(S,H2,L2), sup(R,S).
super(R,S):- sup(R,S), conf_defeasible(R,H1,L1), defeasible(S,H2,L2), sup(R,S).

negation(~(X),X):- !.
negation(X,~(X)).

append([],List,List).
append([Head|Tail],List2,[Head|Result]):- append(Tail,List2,Result).

member(N,[N|Tail]).
member(N,[I|Tail]):- member(N,Tail).

minus_set([E|X],Y,Z):- member(E,Y),minus_set(X,Y,Z),!.
minus_set([E|X],Y,[E|Z]):- minus_set(X,Y,Z),sk_not(member(E,Y)), !.
minus_set([],Y,[]).

strict(e,w,r).
defeasible(y,t,e).
fact(w).
sup(e,w).


